/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001, 2002 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 * "$Id:"
 */

#ifndef TVMET_XPR_OPERATORS_H
#define TVMET_XPR_OPERATORS_H

#include <tvmet/Functional.h>

namespace tvmet {

/***********************************************************************
 * global binary math, bitops and logical operators
 ***********************************************************************/

#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)                   	\
template<class E1, class E2, std::size_t Sz>                   		\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                    			      			\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprVector<E1, Sz>,              			      		\
    XprVector<E2, Sz>               			      		\
  >,                   			      			      	\
  Sz                   			      			      	\
>                   			      		      		\
operator OP (XprVector<E1, Sz> lhs, XprVector<E2, Sz> rhs) {		\
  typedef XprBinOp<							\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprVector<E1, Sz>, 							\
    XprVector<E2, Sz>               			      		\
  >		    					ExprType; 	\
  return XprVector<ExprType, Sz>(ExprType(lhs, rhs));			\
}                   			      		      		\
                   			      		      		\
template<class E1, class E2, std::size_t Rows, std::size_t Cols>	\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                    			      			\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows, Cols>,         			      		\
    XprMatrix<E2, Rows, Cols>         			      		\
  >,                   			      			      	\
  Rows, Cols           			      			      	\
>                   			      		      		\
operator OP (XprMatrix<E1, Rows, Cols> lhs, XprMatrix<E2, Rows, Cols> rhs) { \
  typedef XprBinOp<							\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows, Cols>,						\
    XprMatrix<E2, Rows, Cols>						\
  >		    					ExprType; 	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs, rhs));		\
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

/*
 * XprVector<E, Sz> and POD's
 * XprMatrix<E, Rows, Cols> and POD's
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP, TP)              	\
template<class E, std::size_t Sz>          				\
inline                 			      			      	\
XprVector<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, TP >,      			\
    XprVector<E, Sz>,			          			\
    XprLiteral< TP >          						\
  >,          								\
  Sz          								\
>          								\
operator OP (XprVector<E, Sz> lhs, TP rhs) {				\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, TP >,   				\
    XprVector<E, Sz>,			        			\
    XprLiteral< TP >              			      		\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(lhs, XprLiteral< TP >(rhs))); \
}          								\
          								\
template<class E, std::size_t Sz>   					\
inline                 			      			      	\
XprVector<          							\
  XprBinOp<          							\
    FUNCTIONAL<TP, typename E::value_type>,    				\
    XprLiteral< TP >,          						\
    XprVector<E, Sz>   				      			\
  >,          								\
  Sz          								\
>          								\
operator OP (TP lhs, XprVector<E, Sz> rhs) {         			\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, typename E::value_type>,   				\
    XprLiteral< TP >,         						\
    XprVector<E, Sz>           			      			\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(XprLiteral< TP >(lhs), rhs)); \
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, TP >,      			\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< TP >          						\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (XprMatrix<E, Rows, Cols> lhs, TP rhs) {			\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, TP >,   				\
    XprMatrix<E, Rows, Cols>,		        			\
    XprLiteral< TP >              			      		\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs, XprLiteral< TP >(rhs))); \
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL<TP, typename E::value_type>,    				\
    XprLiteral< TP >,          						\
    XprMatrix<E, Rows, Cols>			      			\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (TP lhs, XprMatrix<E, Rows, Cols> rhs) {  			\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, typename E::value_type>,   				\
    XprLiteral< TP >,         						\
    XprMatrix<E, Rows, Cols>   			      			\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(XprLiteral< TP >(lhs), rhs)); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +, int)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, int)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, int)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, int)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |, int)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<, int)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>, int)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, int)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, int)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, int)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, int)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, int)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, int)

DECLARE_BINARY_OPERATOR(fcnl_Add, +, float)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, float)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, float)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, float)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, float)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, float)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, float)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, float)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, float)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, float)

DECLARE_BINARY_OPERATOR(fcnl_Add, +, double)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, double)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, double)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, double)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, double)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, double)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, double)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, double)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, double)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, double)

#ifdef HAVE_LONG_DOUBLE
DECLARE_BINARY_OPERATOR(fcnl_Add, +, long double)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, long double)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, long double)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, long double)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, long double)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, long double)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, long double)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, long double)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, long double)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, long double)
#endif // HAVE_LONG_DOUBLE

#undef DECLARE_BINARY_OPERATOR

#ifdef HAVE_COMPLEX
/*
 * XprVector<E, Sz> and std::complex<>
 * XprMatrix<E, Rows, Cols> and std::complex<>
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP, TP)              	\
template<class E, std::size_t Sz, class T>     				\
inline                 			      			      	\
XprVector<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, std::complex<T> >, 		\
    XprVector<E, Sz>,			          			\
    XprLiteral< std::complex<T> > 					\
  >,          								\
  Sz          								\
>          								\
operator OP (XprVector<E, Sz> lhs, const std::complex<T>& rhs) {	\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, std::complex<T> >, 		\
    XprVector<E, Sz>,			        			\
    XprLiteral< std::complex<T> >              		      		\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(lhs, XprLiteral< std::complex<T> >(rhs))); \
}          								\
          								\
template<class E, std::size_t Sz, class T> 				\
inline                 			      			      	\
XprVector<          							\
  XprBinOp<          							\
    FUNCTIONAL<std::complex<T>, typename E::value_type>, 		\
    XprLiteral< std::complex<T> >,          				\
    XprVector<E, Sz>   				      			\
  >,          								\
  Sz          								\
>          								\
operator OP (const std::complex<T>& lhs, XprVector<E, Sz> rhs) {	\
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, typename E::value_type>,  		\
    XprLiteral< std::complex<T> >,         				\
    XprVector<E, Sz>           			      			\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(XprLiteral< std::complex<T> >(lhs), rhs)); \
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, std::complex<T> >,      		\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (XprMatrix<E, Rows, Cols> lhs, const std::complex<T>& rhs) { \
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, std::complex<T> >,   		\
    XprMatrix<E, Rows, Cols>,		        			\
    XprLiteral< std::complex<T> >              			      	\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs, XprLiteral< std::complex<T> >(rhs))); \
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL<std::complex<T>, typename E::value_type>,    		\
    XprLiteral< std::complex<T> >,          				\
    XprMatrix<E, Rows, Cols>			      			\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const std::complex<T>& lhs, XprMatrix<E, Rows, Cols> rhs) { \
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, typename E::value_type>,   		\
    XprLiteral< std::complex<T> >,         				\
    XprMatrix<E, Rows, Cols>   			      			\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(XprLiteral< std::complex<T> >(lhs), rhs)); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, std::complex<T>)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, std::complex<T>)
#undef DECLARE_BINARY_OPERATOR

#endif // HAVE_COMPLEX

/***********************************************************************
 * global unary operators
 ***********************************************************************/

/*
 * Binary Operator on XprVector<E, Sz>
 * Operator on XprMatrix<E, Rows, Cols>
 */
#define DECLARE_UNARY_OPERATOR(FUNCTIONAL, OP)                   	\
template <class E, std::size_t Sz>					\
inline                                                        		\
XprVector<                                                        	\
  XprUnOp<                                                        	\
    FUNCTIONAL<typename E::value_type>,                                 \
    XprVector<E, Sz>							\
  >,                                                        		\
  Sz									\
>                                                        		\
operator OP (XprVector<E, Sz> rhs) {					\
  typedef XprUnOp<                                                      \
    FUNCTIONAL<typename E::value_type>,                                 \
    XprVector<E, Sz>							\
  >  							 ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(rhs));			\
}                                                        		\
                                                        		\
template <class E, std::size_t Rows, std::size_t Cols>			\
inline                                                        		\
XprMatrix<                                                        	\
  XprUnOp<                                                        	\
    FUNCTIONAL<typename E::value_type>,                                 \
    XprMatrix<E, Rows, Cols>						\
  >,                                                        		\
  Rows, Cols								\
>                                                        		\
operator OP (XprMatrix<E, Rows, Cols> rhs) {				\
  typedef XprUnOp<                                                      \
    FUNCTIONAL<typename E::value_type>,                                 \
    XprMatrix<E, Rows, Cols>						\
  >  							 ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(rhs));		\
}

DECLARE_UNARY_OPERATOR(fcnl_LogicalNot, !)
DECLARE_UNARY_OPERATOR(fcnl_BitwiseNot, ~)
DECLARE_UNARY_OPERATOR(fcnl_Negate, -)
#undef DECLARE_UNARY_OPERATOR

} // namespace tvmet

#endif // TVMET_XPR_OPERATORS_H

// Local Variables:
// mode:C++
// End:
